@@ -5,6 +5,11 @@
 
 #include "controller.h"
 
+#include <connection.h>
+#include <csapi/content-repo.h>
+
+using namespace Quotient;
+
 LinkPreviewer::LinkPreviewer(QObject *parent)
     : QObject(parent)
     , m_loaded(false)
@@ -51,8 +56,8 @@ void LinkPreviewer::setUrl(QUrl url)
 
         connect(job, &BaseJob::success, this, [this, job]() {
             const auto json = job->jsonData();
-            m_title = json["og:title"].toString();
-            m_description = json["og:description"].toString();
+            m_title = json["og:title"].toString().trimmed();
+            m_description = json["og:description"].toString().trimmed().replace("\n", " ");
             m_imageSource = json["og:image"].toString();
             m_loaded = true;
             Q_EMIT titleChanged();
