@@ -1,4 +1,4 @@
-// SPDX-FileCopyrightText: 2022 Tobias Fella <fella@posteo.de>
+// SPDX-FileCopyrightText: 2022 Tobias Fella <tobias.fella@kde.org>
 // SPDX-License-Identifier: LGPL-2.0-or-later
 
 #include "completionmodel.h"
@@ -16,7 +16,7 @@ CompletionModel::CompletionModel(QObject *parent)
     : QAbstractListModel(parent)
     , m_filterModel(new CompletionProxyModel())
     , m_userListModel(new UserListModel(this))
-    , m_emojiModel(new KConcatenateRowsProxyModel(this))
+    , m_emojiModel(new QConcatenateTablesProxyModel(this))
 {
     connect(this, &CompletionModel::textChanged, this, &CompletionModel::updateCompletion);
     connect(this, &CompletionModel::roomChanged, this, [this]() {
@@ -145,7 +145,7 @@ void CompletionModel::updateCompletion()
         m_filterModel->setFullText(m_fullText);
         m_filterModel->setFilterText(m_text);
         m_filterModel->invalidate();
-    } else if (text().startsWith(QLatin1Char(':'))
+    } else if (text().startsWith(QLatin1Char(':')) && !text()[1].isUpper()
                && (m_fullText.indexOf(QLatin1Char(':'), 1) == -1
                    || (m_fullText.indexOf(QLatin1Char(' ')) != -1 && m_fullText.indexOf(QLatin1Char(':'), 1) > m_fullText.indexOf(QLatin1Char(' '), 1)))) {
         m_filterModel->setSourceModel(m_emojiModel);
