@@ -27,7 +27,6 @@
 #include <KLocalizedString>
 
 #include "neochatuser.h"
-#include "utils.h"
 
 using namespace Quotient;
 
@@ -54,6 +53,9 @@ QHash<int, QByteArray> MessageEventModel::roleNames() const
     roles[UserMarkerRole] = "userMarker";
     roles[ShowAuthorRole] = "showAuthor";
     roles[ShowSectionRole] = "showSection";
+    roles[ReadMarkersRole] = "readMarkers";
+    roles[ReadMarkersStringRole] = "readMarkersString";
+    roles[ShowReadMarkersRole] = "showReadMarkers";
     roles[ReactionRole] = "reaction";
     roles[IsEditedRole] = "isEdited";
     roles[SourceRole] = "source";
@@ -64,9 +66,9 @@ QHash<int, QByteArray> MessageEventModel::roleNames() const
     roles[VerifiedRole] = "verified";
     roles[DisplayNameForInitialsRole] = "displayNameForInitials";
     roles[AuthorDisplayNameRole] = "authorDisplayName";
-    roles[IsNameChangeRole] = "isNameChange";
-    roles[IsAvatarChangeRole] = "isAvatarChange";
     roles[IsRedactedRole] = "isRedacted";
+    roles[GenericDisplayRole] = "genericDisplay";
+    roles[IsPendingRole] = "isPending";
     return roles;
 }
 
@@ -175,6 +177,7 @@ void MessageEventModel::setRoom(NeoChatRoom *room)
         });
         connect(m_currentRoom, &Room::pendingEventAdded, this, &MessageEventModel::endInsertRows);
         connect(m_currentRoom, &Room::pendingEventAboutToMerge, this, [this](RoomEvent *, int i) {
+            Q_EMIT dataChanged(index(i, 0), index(i, 0), {IsPendingRole});
             if (i == 0) {
                 return; // No need to move anything, just refresh
             }
@@ -213,6 +216,12 @@ void MessageEventModel::setRoom(NeoChatRoom *room)
             }
             refreshEventRoles(eventId, {ReactionRole, Qt::DisplayRole});
         });
+        connect(m_currentRoom, &Room::changed, this, [this]() {
+            for (auto it = m_currentRoom->messageEvents().rbegin(); it != m_currentRoom->messageEvents().rend(); ++it) {
+                auto event = it->event();
+                refreshEventRoles(event->id(), {ReadMarkersRole, ReadMarkersStringRole});
+            }
+        });
         connect(m_currentRoom, &Room::newFileTransfer, this, &MessageEventModel::refreshEvent);
         connect(m_currentRoom, &Room::fileTransferProgress, this, &MessageEventModel::refreshEvent);
         connect(m_currentRoom, &Room::fileTransferCompleted, this, &MessageEventModel::refreshEvent);
@@ -462,6 +471,14 @@ QVariant MessageEventModel::data(const QModelIndex &idx, int role) const
         return m_currentRoom->eventToString(evt, Qt::RichText);
     }
 
+    if (role == GenericDisplayRole) {
+        if (evt.isRedacted()) {
+            return i18n("<i>[This message was deleted]</i>");
+        }
+
+        return m_currentRoom->eventToGenericString(evt);
+    }
+
     if (role == FormattedBodyRole) {
         if (auto e = eventCast<const RoomMessageEvent>(&evt)) {
             if (e->hasTextContent() && e->mimeType().name() != "text/plain") {
@@ -560,7 +577,7 @@ QVariant MessageEventModel::data(const QModelIndex &idx, int role) const
     }
 
     if (role == HighlightRole) {
-        return m_currentRoom->isEventHighlighted(&evt);
+        return !m_currentRoom->isDirectChat() && m_currentRoom->isEventHighlighted(&evt);
     }
 
     if (role == FileMimetypeIcon) {
@@ -591,9 +608,17 @@ QVariant MessageEventModel::data(const QModelIndex &idx, int role) const
             return pendingIt->deliveryStatus();
         }
 
-        auto *memberEvent = timelineIt->viewAs<RoomMemberEvent>();
-        if (memberEvent) {
-            if ((memberEvent->isJoin() || memberEvent->isLeave()) && !NeoChatConfig::self()->showLeaveJoinEvent()) {
+        if (evt.isStateEvent() && !NeoChatConfig::self()->showStateEvent()) {
+            return EventStatus::Hidden;
+        }
+
+        if (auto roomMemberEvent = eventCast<const RoomMemberEvent>(&evt)) {
+            if ((roomMemberEvent->isJoin() || roomMemberEvent->isLeave()) && !NeoChatConfig::self()->showLeaveJoinEvent()) {
+                return EventStatus::Hidden;
+            } else if (roomMemberEvent->isRename() && !roomMemberEvent->isJoin() && !roomMemberEvent->isLeave() && !NeoChatConfig::self()->showRename()) {
+                return EventStatus::Hidden;
+            } else if (roomMemberEvent->isAvatarUpdate() && !roomMemberEvent->isJoin() && !roomMemberEvent->isLeave()
+                       && !NeoChatConfig::self()->showAvatarUpdate()) {
                 return EventStatus::Hidden;
             }
         }
@@ -780,6 +805,65 @@ QVariant MessageEventModel::data(const QModelIndex &idx, int role) const
         return false;
     }
 
+    if (role == ReadMarkersRole) {
+#ifdef QUOTIENT_07
+        auto userIds = room()->userIdsAtEvent(evt.id());
+        userIds.remove(m_currentRoom->localUser()->id());
+#else
+        auto userIds = room()->usersAtEventId(evt.id());
+        userIds.removeAll(m_currentRoom->localUser());
+#endif
+
+        QVariantList users;
+        users.reserve(userIds.size());
+        for (const auto &userId : userIds) {
+#ifdef QUOTIENT_07
+            auto user = static_cast<NeoChatUser *>(m_currentRoom->user(userId));
+#else
+            auto user = static_cast<NeoChatUser *>(userId);
+#endif
+            users += userAtEvent(user, m_currentRoom, evt);
+        }
+
+        return users;
+    }
+
+    if (role == ReadMarkersStringRole) {
+#ifdef QUOTIENT_07
+        auto userIds = room()->userIdsAtEvent(evt.id());
+        userIds.remove(m_currentRoom->localUser()->id());
+#else
+        auto userIds = room()->usersAtEventId(evt.id());
+        userIds.removeAll(m_currentRoom->localUser());
+#endif
+        /**
+         * The string ends up in the form
+         * "x users: user1DisplayName, user2DisplayName, etc."
+         */
+        QString readMarkersString = i18np("1 user: ", "%1 users: ", userIds.size());
+        for (const auto &userId : userIds) {
+#ifdef QUOTIENT_07
+            auto user = static_cast<NeoChatUser *>(m_currentRoom->user(userId));
+#else
+            auto user = static_cast<NeoChatUser *>(userId);
+#endif
+            readMarkersString += user->displayname(m_currentRoom) + i18nc("list separator", ", ");
+        }
+        readMarkersString.chop(2);
+        return readMarkersString;
+    }
+
+    if (role == ShowReadMarkersRole) {
+#ifdef QUOTIENT_07
+        auto userIds = room()->userIdsAtEvent(evt.id());
+        userIds.remove(m_currentRoom->localUser()->id());
+#else
+        auto userIds = room()->usersAtEventId(evt.id());
+        userIds.removeAll(m_currentRoom->localUser());
+#endif
+        return userIds.size() > 0;
+    }
+
     if (role == ReactionRole) {
         const auto &annotations = m_currentRoom->relatedEvents(evt, EventRelation::Annotation());
         if (annotations.isEmpty()) {
@@ -880,25 +964,14 @@ QVariant MessageEventModel::data(const QModelIndex &idx, int role) const
         }
     }
 
-    if (role == IsNameChangeRole) {
-        auto roomMemberEvent = eventCast<const RoomMemberEvent>(&evt);
-        if (roomMemberEvent) {
-            return roomMemberEvent->isRename();
-        }
-        return false;
-    }
-
-    if (role == IsAvatarChangeRole) {
-        auto roomMemberEvent = eventCast<const RoomMemberEvent>(&evt);
-        if (roomMemberEvent) {
-            return roomMemberEvent->isAvatarUpdate();
-        }
-        return false;
-    }
     if (role == IsRedactedRole) {
         return evt.isRedacted();
     }
 
+    if (role == IsPendingRole) {
+        return row < m_currentRoom->pendingEvents().size();
+    }
+
     return {};
 }
 
@@ -966,6 +1039,9 @@ QVariant MessageEventModel::getLatestMessageFromIndex(const int baseline)
     for (auto it = timelineBottom; it != limit; ++it) {
         auto evt = it->event();
         auto e = eventCast<const RoomMessageEvent>(evt);
+        if (!e) {
+            continue;
+        }
 
         auto content = (*it)->contentJson();
 
