@@ -13,9 +13,11 @@
 #include <QMediaMetaData>
 #include <QMediaPlayer>
 
+#include <jobs/basejob.h>
 #include <qcoro/qcorosignal.h>
 
 #include <connection.h>
+#include <csapi/account-data.h>
 #include <csapi/directory.h>
 #include <csapi/pushrules.h>
 #include <csapi/redaction.h>
@@ -47,7 +49,7 @@
 #endif
 #include "filetransferpseudojob.h"
 #include "stickerevent.h"
-#include "utils.h"
+#include "texthandler.h"
 
 #ifndef Q_OS_ANDROID
 #include <KIO/Job>
@@ -98,6 +100,14 @@ NeoChatRoom::NeoChatRoom(Connection *connection, QString roomId, JoinState joinS
         Q_EMIT canEncryptRoomChanged();
     });
     connect(connection, &Connection::capabilitiesLoaded, this, &NeoChatRoom::maxRoomVersionChanged);
+    connect(this, &Room::changed, this, [this]() {
+        Q_EMIT defaultUrlPreviewStateChanged();
+    });
+    connect(this, &Room::accountDataChanged, this, [this](QString type) {
+        if (type == "org.matrix.room.preview_urls") {
+            Q_EMIT urlPreviewEnabledChanged();
+        }
+    });
 }
 
 void NeoChatRoom::uploadFile(const QUrl &url, const QString &body)
@@ -205,7 +215,7 @@ void NeoChatRoom::sendTypingNotification(bool isTyping)
     connection()->callApi<SetTypingJob>(BackgroundRequest, localUser()->id(), id(), isTyping, 10000);
 }
 
-const RoomEvent *NeoChatRoom::lastEvent(bool ignoreStateEvent) const
+const RoomEvent *NeoChatRoom::lastEvent() const
 {
     for (auto timelineItem = messageEvents().rbegin(); timelineItem < messageEvents().rend(); timelineItem++) {
         const RoomEvent *event = timelineItem->get();
@@ -217,8 +227,21 @@ const RoomEvent *NeoChatRoom::lastEvent(bool ignoreStateEvent) const
             continue;
         }
 
-        if (event->isStateEvent()
-            && (ignoreStateEvent || !NeoChatConfig::self()->showLeaveJoinEvent() || static_cast<const StateEventBase &>(*event).repeatsState())) {
+        if (event->isStateEvent() && !NeoChatConfig::self()->showStateEvent()) {
+            continue;
+        }
+
+        if (auto roomMemberEvent = eventCast<const RoomMemberEvent>(event)) {
+            if ((roomMemberEvent->isJoin() || roomMemberEvent->isLeave()) && !NeoChatConfig::self()->showLeaveJoinEvent()) {
+                continue;
+            } else if (roomMemberEvent->isRename() && !roomMemberEvent->isJoin() && !roomMemberEvent->isLeave() && !NeoChatConfig::self()->showRename()) {
+                continue;
+            } else if (roomMemberEvent->isAvatarUpdate() && !roomMemberEvent->isJoin() && !roomMemberEvent->isLeave()
+                       && !NeoChatConfig::self()->showAvatarUpdate()) {
+                continue;
+            }
+        }
+        if (event->isStateEvent() && static_cast<const StateEventBase &>(*event).repeatsState()) {
             continue;
         }
 
@@ -232,6 +255,14 @@ const RoomEvent *NeoChatRoom::lastEvent(bool ignoreStateEvent) const
             continue;
         }
 
+#ifdef QUOTIENT_07
+        if (auto lastEvent = eventCast<const StateEvent>(event)) {
+#else
+        if (auto lastEvent = eventCast<const StateEventBase>(event)) {
+#endif
+            return lastEvent;
+        }
+
         if (auto lastEvent = eventCast<const RoomMessageEvent>(event)) {
             return lastEvent;
         }
@@ -257,10 +288,11 @@ bool NeoChatRoom::lastEventIsSpoiler() const
     return false;
 }
 
-QString NeoChatRoom::lastEventToString() const
+QString NeoChatRoom::lastEventToString(Qt::TextFormat format, bool stripNewlines) const
 {
     if (auto event = lastEvent()) {
-        return roomMembername(event->senderId()) + (event->isStateEvent() ? " " : ": ") + eventToString(*event);
+        return roomMembername(event->senderId()) + (event->isStateEvent() ? QLatin1String(" ") : QLatin1String(": "))
+            + eventToString(*event, format, stripNewlines);
     }
     return QLatin1String("");
 }
@@ -321,7 +353,7 @@ QDateTime NeoChatRoom::lastActiveTime()
         return QDateTime();
     }
 
-    if (auto event = lastEvent(true)) {
+    if (auto event = lastEvent()) {
         return event->originTimestamp();
     }
 
@@ -329,45 +361,6 @@ QDateTime NeoChatRoom::lastActiveTime()
     return messageEvents().rbegin()->get()->originTimestamp();
 }
 
-QString NeoChatRoom::subtitleText()
-{
-    static const QRegularExpression blockquote("(\r\n\t|\n|\r\t|)> ");
-    static const QRegularExpression heading("(\r\n\t|\n|\r\t|)\\#{1,6} ");
-    static const QRegularExpression newlines("(\r\n\t|\n|\r\t|\r\n)");
-    static const QRegularExpression bold1("(\\*\\*|__)(?=\\S)([^\\r]*\\S)\\1");
-    static const QRegularExpression bold2("(\\*|_)(?=\\S)([^\\r]*\\S)\\1");
-    static const QRegularExpression strike1("~~(.*)~~");
-    static const QRegularExpression strike2("~(.*)~");
-    static const QRegularExpression del("<del>(.*)</del>");
-    static const QRegularExpression multileLineCode("```([^```]+)```");
-    static const QRegularExpression singleLinecode("`([^`]+)`");
-    QString subtitle = lastEventToString().size() == 0 ? topic() : lastEventToString();
-
-    subtitle
-        // replace blockquote, i.e. '> text'
-        .replace(blockquote, " ")
-        // replace headings, i.e. "# text"
-        .replace(heading, " ")
-        // replace newlines
-        .replace(newlines, " ")
-        // replace '**text**' and '__text__'
-        .replace(bold1, "\\2")
-        // replace '*text*' and '_text_'
-        .replace(bold2, "\\2")
-        // replace '~~text~~'
-        .replace(strike1, "\\1")
-        // replace '~text~'
-        .replace(strike2, "\\1")
-        // replace '<del>text</del>'
-        .replace(del, "\\1")
-        // replace '```code```'
-        .replace(multileLineCode, "\\1")
-        // replace '`code`'
-        .replace(singleLinecode, "\\1");
-
-    return subtitle.size() > 0 ? subtitle : QStringLiteral(" ");
-}
-
 int NeoChatRoom::savedTopVisibleIndex() const
 {
     return firstDisplayedMarker() == historyEdge() ? 0 : int(firstDisplayedMarker() - messageEvents().rbegin());
@@ -451,7 +444,7 @@ QString NeoChatRoom::avatarMediaId() const
     return {};
 }
 
-QString NeoChatRoom::eventToString(const RoomEvent &evt, Qt::TextFormat format, bool removeReply) const
+QString NeoChatRoom::eventToString(const RoomEvent &evt, Qt::TextFormat format, bool stripNewlines) const
 {
     const bool prettyPrint = (format == Qt::RichText);
 
@@ -462,58 +455,48 @@ QString NeoChatRoom::eventToString(const RoomEvent &evt, Qt::TextFormat format,
     return visit(
 #endif
         evt,
-        [this, prettyPrint, removeReply](const RoomMessageEvent &e) {
+        [this, format, stripNewlines](const RoomMessageEvent &e) {
             using namespace MessageEventContent;
 
-            // 1. prettyPrint/HTML
-            if (prettyPrint && e.hasTextContent() && e.mimeType().name() != "text/plain") {
-                auto htmlBody = static_cast<const TextContent *>(e.content())->body;
-                if (removeReply) {
-                    htmlBody.remove(utils::removeRichReplyRegex);
-                }
-                htmlBody.replace(utils::userPillRegExp, R"(<b class="user-pill">\1</b>)");
-                htmlBody.replace(utils::strikethroughRegExp, "<s>\\1</s>");
-
-                auto url = connection()->homeserver();
-                auto base = url.scheme() + QStringLiteral("://") + url.host() + (url.port() != -1 ? ':' + QString::number(url.port()) : QString());
-                htmlBody.replace(utils::mxcImageRegExp, QStringLiteral(R"(<img \1 src="%1/_matrix/media/r0/download/\2/\3" \4 > )").arg(base));
-
-                return htmlBody;
-            }
+            TextHandler textHandler;
 
             if (e.hasFileContent()) {
-                auto fileCaption = e.content()->fileInfo()->originalName.toHtmlEscaped();
+                auto fileCaption = e.content()->fileInfo()->originalName;
                 if (fileCaption.isEmpty()) {
-                    fileCaption = prettyPrint ? Quotient::prettyPrint(e.plainBody()) : e.plainBody();
+                    fileCaption = e.plainBody();
                 } else if (e.content()->fileInfo()->originalName != e.plainBody()) {
                     fileCaption = e.plainBody() + " | " + fileCaption;
                 }
-                return !fileCaption.isEmpty() ? fileCaption : i18n("a file");
+                textHandler.setData(fileCaption);
+                return !fileCaption.isEmpty() ? textHandler.handleRecievePlainText() : i18n("a file");
             }
 
-            // 2. prettyPrint/text 3. plainText/HTML 4. plainText/text
-            QString plainBody;
-            if (e.hasTextContent() && e.content() && e.mimeType().name() == "text/plain") { // 2/4
-                plainBody = static_cast<const TextContent *>(e.content())->body;
-            } else { // 3
-                plainBody = e.plainBody();
+            QString body;
+            if (e.hasTextContent() && e.content()) {
+                body = static_cast<const TextContent *>(e.content())->body;
+            } else {
+                body = e.plainBody();
             }
 
-            if (prettyPrint) {
-                if (removeReply) {
-                    plainBody.remove(utils::removeReplyRegex);
-                }
-                return Quotient::prettyPrint(plainBody);
+            textHandler.setData(body);
+
+            Qt::TextFormat inputFormat;
+            if (e.mimeType().name() == "text/plain") {
+                inputFormat = Qt::PlainText;
+            } else {
+                inputFormat = Qt::RichText;
             }
-            if (removeReply) {
-                return plainBody.remove(utils::removeReplyRegex);
+
+            if (format == Qt::RichText) {
+                return textHandler.handleRecieveRichText(inputFormat, this, &e, stripNewlines);
+            } else {
+                return textHandler.handleRecievePlainText(inputFormat, stripNewlines);
             }
-            return plainBody;
         },
         [](const StickerEvent &e) {
             return e.body();
         },
-        [this](const RoomMemberEvent &e) {
+        [this, prettyPrint](const RoomMemberEvent &e) {
             // FIXME: Rewind to the name that was at the time of this event
             auto subjectName = this->htmlSafeMemberName(e.userId());
             if (e.membership() == MembershipType::Leave) {
@@ -526,8 +509,11 @@ QString NeoChatRoom::eventToString(const RoomEvent &evt, Qt::TextFormat format,
 #endif
                 }
             }
-            subjectName = QStringLiteral("<a href=\"https://matrix.to/#/%1\" style=\"color: %2\">%3</a>")
-                              .arg(e.userId(), static_cast<NeoChatUser *>(user(e.userId()))->color().name(), subjectName);
+
+            if (prettyPrint) {
+                subjectName = QStringLiteral("<a href=\"https://matrix.to/#/%1\" style=\"color: %2\">%3</a>")
+                                  .arg(e.userId(), static_cast<NeoChatUser *>(user(e.userId()))->color().name(), subjectName);
+            }
 
             // The below code assumes senderName output in AuthorRole
             switch (e.membership()) {
@@ -658,6 +644,135 @@ QString NeoChatRoom::eventToString(const RoomEvent &evt, Qt::TextFormat format,
         i18n("Unknown event"));
 }
 
+QString NeoChatRoom::eventToGenericString(const RoomEvent &evt) const
+{
+#ifdef QUOTIENT_07
+    return switchOnType(
+#else
+    return visit(
+#endif
+        evt,
+        [](const RoomMessageEvent &e) {
+            Q_UNUSED(e)
+            return i18n("sent a message");
+        },
+        [](const StickerEvent &e) {
+            Q_UNUSED(e)
+            return i18n("sent a sticker");
+        },
+        [](const RoomMemberEvent &e) {
+            switch (e.membership()) {
+            case MembershipType::Invite:
+                if (e.repeatsState()) {
+                    return i18n("reinvited someone to the room");
+                }
+                Q_FALLTHROUGH();
+            case MembershipType::Join: {
+                QString text{};
+                // Part 1: invites and joins
+                if (e.repeatsState()) {
+                    text = i18n("joined the room (repeated)");
+                } else if (e.changesMembership()) {
+                    text = e.membership() == MembershipType::Invite ? i18n("invited someone to the room") : i18n("joined the room");
+                }
+                if (!text.isEmpty()) {
+                    return text;
+                }
+                // Part 2: profile changes of joined members
+                if (e.isRename()) {
+                    if (e.displayName().isEmpty()) {
+                        text = i18nc("their refers to a singular user", "cleared their display name");
+                    } else {
+                        text = i18nc("their refers to a singular user", "changed their display name");
+                    }
+                }
+                if (e.isAvatarUpdate()) {
+                    if (!text.isEmpty()) {
+                        text += i18n(" and ");
+                    }
+                    if (e.avatarUrl().isEmpty()) {
+                        text += i18nc("their refers to a singular user", "cleared their avatar");
+#ifdef QUOTIENT_07
+                    } else if (!e.prevContent()->avatarUrl) {
+#else
+                    } else if (e.prevContent()->avatarUrl.isEmpty()) {
+#endif
+                        text += i18n("set an avatar");
+                    } else {
+                        text += i18nc("their refers to a singular user", "updated their avatar");
+                    }
+                }
+                if (text.isEmpty()) {
+                    text = i18nc("<user> changed nothing", "changed nothing");
+                }
+                return text;
+            }
+            case MembershipType::Leave:
+                if (e.prevContent() && e.prevContent()->membership == MembershipType::Invite) {
+                    return (e.senderId() != e.userId()) ? i18n("withdrew a user's invitation") : i18n("rejected the invitation");
+                }
+
+                if (e.prevContent() && e.prevContent()->membership == MembershipType::Ban) {
+                    return (e.senderId() != e.userId()) ? i18n("unbanned a user") : i18n("self-unbanned");
+                }
+                return (e.senderId() != e.userId()) ? i18n("put a user out of the room") : i18n("left the room");
+            case MembershipType::Ban:
+                if (e.senderId() != e.userId()) {
+                    return i18n("banned a user from the room");
+                } else {
+                    return i18n("self-banned from the room");
+                }
+            case MembershipType::Knock: {
+                return i18n("requested an invite");
+            }
+            default:;
+            }
+            return i18n("made something unknown");
+        },
+        [](const RoomCanonicalAliasEvent &e) {
+            return (e.alias().isEmpty()) ? i18n("cleared the room main alias") : i18n("set the room main alias");
+        },
+        [](const RoomNameEvent &e) {
+            return (e.name().isEmpty()) ? i18n("cleared the room name") : i18n("set the room name");
+        },
+        [](const RoomTopicEvent &e) {
+            return (e.topic().isEmpty()) ? i18n("cleared the topic") : i18n("set the topic");
+        },
+        [](const RoomAvatarEvent &) {
+            return i18n("changed the room avatar");
+        },
+        [](const EncryptionEvent &) {
+            return i18n("activated End-to-End Encryption");
+        },
+        [](const RoomCreateEvent &e) {
+            return e.isUpgrade() ? i18n("upgraded the room version") : i18n("created the room");
+        },
+        [](const RoomPowerLevelsEvent &) {
+            return i18nc("'power level' means permission level", "changed the power levels for this room");
+        },
+        [](const StateEventBase &e) {
+            if (e.matrixType() == QLatin1String("m.room.server_acl")) {
+                return i18n("changed the server access control lists for this room");
+            }
+            if (e.matrixType() == QLatin1String("im.vector.modular.widgets")) {
+                if (e.fullJson()["unsigned"]["prev_content"].toObject().isEmpty()) {
+                    return i18n("added a widget");
+                }
+                if (e.contentJson().isEmpty()) {
+                    return i18n("removed a widget");
+                }
+                return i18n("configured a widget");
+            }
+            return i18n("updated the state");
+        },
+#ifdef QUOTIENT_07
+        [](const PollStartEvent &e) {
+            return i18n("started a poll");
+        },
+#endif
+        i18n("Unknown event"));
+}
+
 void NeoChatRoom::changeAvatar(const QUrl &localFile)
 {
     const auto job = connection()->uploadFile(localFile.toLocalFile());
@@ -856,7 +971,7 @@ bool NeoChatRoom::canSendState(const QString &eventType) const
     auto currentPl = plEvent->powerLevelForUser(localUser()->id());
 
 #ifndef QUOTIENT_07
-    if (eventType == "m.room.history_visibility") {
+    if (eventType == "m.room.history_visibility" || eventType == "org.matrix.room.preview_urls") {
         return false;
     } else {
         return currentPl >= pl;
@@ -946,6 +1061,91 @@ void NeoChatRoom::setHistoryVisibility(const QString &historyVisibilityRule)
     // Not emitting historyVisibilityChanged() here, since that would override the change in the UI with the *current* value, which is not the *new* value.
 }
 
+bool NeoChatRoom::defaultUrlPreviewState() const
+{
+#ifdef QUOTIENT_07
+    auto urlPreviewsDisabled = currentState().get("org.matrix.room.preview_urls");
+#else
+    auto urlPreviewsDisabled = getCurrentState("org.matrix.room.preview_urls");
+#endif
+
+    // Some rooms will not have this state event set so check for a nullptr return.
+    if (urlPreviewsDisabled != nullptr) {
+        return !urlPreviewsDisabled->contentJson()["disable"].toBool();
+    } else {
+        return false;
+    }
+}
+
+void NeoChatRoom::setDefaultUrlPreviewState(const bool &defaultUrlPreviewState)
+{
+    if (!canSendState("org.matrix.room.preview_urls")) {
+        qWarning() << "Power level too low to set the default URL preview state for the room";
+        return;
+    }
+
+    /**
+     * Note the org.matrix.room.preview_urls room state event is completely undocumented
+     * so here it is because I'm nice.
+     *
+     * Also note this is a different event to org.matrix.room.preview_urls for room
+     * account data, because even though it has the same name and content it's totally different.
+     *
+     * {
+     *  "content": {
+     *      "disable": false
+     *  },
+     *  "origin_server_ts": 1673115224071,
+     *  "sender": "@bob:kde.org",
+     *  "state_key": "",
+     *  "type": "org.matrix.room.preview_urls",
+     *  "unsigned": {
+     *      "replaces_state": "replaced_event_id",
+     *      "prev_content": {
+     *          "disable": true
+     *      },
+     *      "prev_sender": "@jeff:kde.org",
+     *      "age": 99
+     *  },
+     *  "event_id": "$event_id",
+     *  "room_id": "!room_id:kde.org"
+     * }
+     *
+     * You just have to set disable to true to disable URL previews by default.
+     */
+#ifdef QUOTIENT_07
+    setState("org.matrix.room.preview_urls", "", QJsonObject{{"disable", !defaultUrlPreviewState}});
+#else
+    qWarning() << "Quotient 0.7 required to set room default url preview setting";
+    return;
+#endif
+}
+
+bool NeoChatRoom::urlPreviewEnabled() const
+{
+    if (hasAccountData("org.matrix.room.preview_urls")) {
+        return !accountData("org.matrix.room.preview_urls")->contentJson()["disable"].toBool();
+    } else {
+        return defaultUrlPreviewState();
+    }
+}
+
+void NeoChatRoom::setUrlPreviewEnabled(const bool &urlPreviewEnabled)
+{
+    /**
+     * Once again this is undocumented and even though the name and content are the
+     * same this is a different event to the org.matrix.room.preview_urls room state event.
+     *
+     * {
+     *  "content": {
+     *      "disable": true
+     *  }
+     *  "type": "org.matrix.room.preview_urls",
+     * }
+     */
+    connection()->callApi<SetAccountDataPerRoomJob>(localUser()->id(), id(), "org.matrix.room.preview_urls", QJsonObject{{"disable", !urlPreviewEnabled}});
+}
+
 void NeoChatRoom::setUserPowerLevel(const QString &userID, const int &powerLevel)
 {
     if (joinedCount() <= 1) {
@@ -985,6 +1185,12 @@ void NeoChatRoom::setUserPowerLevel(const QString &userID, const int &powerLevel
     }
 }
 
+int NeoChatRoom::getUserPowerLevel(const QString &userId) const
+{
+    auto powerLevelEvent = getCurrentState<RoomPowerLevelsEvent>();
+    return powerLevelEvent->powerLevelForUser(userId);
+}
+
 int NeoChatRoom::powerLevel(const QString &eventName, const bool &isStateEvent) const
 {
 #ifdef QUOTIENT_07
@@ -1497,6 +1703,17 @@ void NeoChatRoom::setChatBoxText(const QString &text)
     Q_EMIT chatBoxTextChanged();
 }
 
+QString NeoChatRoom::editText() const
+{
+    return m_editText;
+}
+
+void NeoChatRoom::setEditText(const QString &text)
+{
+    m_editText = text;
+    Q_EMIT editTextChanged();
+}
+
 QString NeoChatRoom::chatBoxReplyId() const
 {
     return m_chatBoxReplyId;
@@ -1573,6 +1790,11 @@ QVector<Mention> *NeoChatRoom::mentions()
     return &m_mentions;
 }
 
+QVector<Mention> *NeoChatRoom::editMentions()
+{
+    return &m_editMentions;
+}
+
 QString NeoChatRoom::savedText() const
 {
     return m_savedText;
@@ -1684,3 +1906,7 @@ int NeoChatRoom::maxRoomVersion() const
     }
     return maxVersion;
 }
+NeoChatUser *NeoChatRoom::directChatRemoteUser() const
+{
+    return dynamic_cast<NeoChatUser *>(connection()->directChatUsers(this)[0]);
+}
